import datetime

from django.contrib.auth import login as auth_login
from django.contrib.auth.hashers import make_password
from django.utils.timezone import localtime
from rest_framework.fields import CharField, SerializerMethodField
from rest_framework.serializers import Serializer
from rest_framework_simplejwt.tokens import SlidingToken

from .schemas.openapi import AutoSchema
from .utils import get_nested_list


class SchemaMixin:
    schema = AutoSchema()


class ModelMixin:

    def get_model(self):
        return self.serializer_class.Meta.model


# Serializer Mixin
class PasswordFieldMixin(Serializer):
    password = CharField(style={"input_type": "password"}, min_length=6, max_length=128, write_only=True)

    # SO https://stackoverflow.com/questions/29746584/django-rest-framework-create-user-with-password
    def validate_password(self, value):
        return make_password(value)


class TokenFieldMixin(Serializer):
    token = SerializerMethodField()

    def get_token(self, obj):
        request = self.context.get("request")
        if request:
            auth_login(request, obj)  # 主要为了记录last_login的, 其他的作用待研究
        else:
            print("request is None, 请在代码手动传入, 否则无法自动登录")
        return str(SlidingToken.for_user(request.user))


class ImagesFieldMixin(Serializer):
    images = SerializerMethodField()

    def get_images(self, obj):
        if hasattr(obj, "images"):
            return get_nested_list([{"uri": item.strip()} for item in obj.images.split("\n") if item.strip()])
        else:
            return "👈⚠️️字段不存在，请去除。"


class TagsFieldMixin(Serializer):
    tags = SerializerMethodField()

    def get_tags(self, obj):
        if hasattr(obj, "tags"):
            return [item.strip() for item in obj.tags.split("#") if item.strip()]
        else:
            return "👈⚠️️字段不存在，请去除。"


class StartDateFieldMixin(Serializer):
    start_date = SerializerMethodField(read_only=True)

    def get_start_date(self, obj):
        return localtime(obj.start_time).strftime("%Y-%m-%d")


class EndDateFieldMixin(Serializer):
    end_date = SerializerMethodField(read_only=True)

    def get_end_date(self, obj):
        if obj.period <= 0:
            return ""
        else:
            return self.get_date_after_period(localtime(obj.start_time), obj.period)

    def get_date_after_period(self, date, period):
        days = period
        if period >= 0:
            days = period - 1
        date_time = date + datetime.timedelta(days=days)
        return date_time.strftime("%Y-%m-%d")


# View Mixin
class OwnerMixin:

    def pre_save(self, obj):
        obj.user_id = self.request.user.id



import os
from urllib.parse import unquote_plus

from django.conf import settings
from django.forms import widgets
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe


class CloudFileWidget(widgets.TextInput):

    class Media:
        js = ["s3direct/dist/index.js"]
        css = {"all": ["s3direct/dist/index.css"]}

    def __init__(self, *args, **kwargs):
        self.dest = kwargs.pop("dest", None)
        super().__init__(*args, **kwargs)

    def render(self, name, value, **kwargs):
        file_url = value or ""
        csrf_cookie_name = getattr(settings, "CSRF_COOKIE_NAME", "csrftoken")
        ctx = {
            "policy_url": reverse("upload-params", args=["aliyun"]),
            "signing_url": reverse("s3direct-signing"),
            "dest": self.dest,
            "name": name,
            "csrf_cookie_name": csrf_cookie_name,
            "file_url": file_url,
            "file_name": os.path.basename(unquote_plus(file_url))
        }
        return mark_safe(render_to_string(os.path.join("s3direct", "s3direct-widget.tpl"), ctx))



from django.contrib.auth.models import AbstractUser
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..utils.translation import dj_gettext, format_gettext_lazy as _f
from .utils import db_master, generate_shard_id, get_user_id


class DateTimeModelMixin(models.Model):
    created_time = models.DateTimeField(_("created_time"), auto_now_add=True)
    updated_time = models.DateTimeField(_("updated_time"), auto_now=True)

    class Meta:
        abstract = True


class AbstractBaseModel(DateTimeModelMixin):
    is_active = models.BooleanField(dj_gettext("active"), default=True)

    # 只是为了PyLint不警告, SO https://stackoverflow.com/questions/45135263/class-has-no-objects-member#45150811
    objects = models.Manager()

    class Meta:
        abstract = True
        ordering = ["-pk"]


class BaseModel(AbstractBaseModel):

    class Meta(AbstractBaseModel.Meta):
        abstract = True


class OwnerModel(BaseModel):
    user_id = models.PositiveIntegerField(_f("user", "id"),
                                          validators=[MinValueValidator(1)],
                                          default=1,
                                          editable=False)
    username = models.CharField(dj_gettext("username"),
                                max_length=150,
                                validators=[AbstractUser.username_validator],
                                editable=False)  # 长度和Django的User保持一致
    user_avatar = models.URLField(_("avatar"), max_length=255, blank=True, editable=False)

    class Meta(BaseModel.Meta):
        abstract = True


class PostModel(OwnerModel):
    """内容发布类模型"""
    title = models.CharField(_("title"), max_length=255)
    summary = models.CharField(_("summary"), max_length=2000, blank=True)
    images = models.CharField(_("images"), max_length=2000, blank=True)
    tags = models.CharField(_("tags"), max_length=255, blank=True)

    class Meta(OwnerModel.Meta):
        abstract = True

    def __str__(self):
        return self.title


class ShardModel(OwnerModel):
    id = models.BigAutoField(primary_key=True)

    class Meta(OwnerModel.Meta):
        abstract = True

    def save(self, using="default", *args, **kwargs):
        self.full_clean()
        self.pk = self.pk or generate_shard_id(self.user_id)
        super().save(using=db_master(self.user_id), *args, **kwargs)


class ShardLCModel(OwnerModel):
    id = models.BigAutoField(primary_key=True)
    is_origin = models.BooleanField(default=True)

    class Meta(OwnerModel.Meta):
        abstract = True

    def save(self, using="default", *args, **kwargs):
        self.full_clean()
        self.pk = self.pk or generate_shard_id(self.user_id)
        super().save(using=db_master(self.user_id), *args, **kwargs)  # 保存第一份
        card_user_id = get_user_id(self.card_id)
        if db_master(card_user_id) != db_master(self.user_id):  # 保存第二份
            self.is_origin = False
            super().save(using=db_master(card_user_id), *args, **kwargs)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend


# SO https://stackoverflow.com/questions/37332190/django-login-with-email#42848959
class ModelAndEmailBackend(ModelBackend):

    def authenticate(self, request, username=None, password=None, **kwargs):
        UserModel = get_user_model()
        if username is None:
            username = kwargs.get(UserModel.USERNAME_FIELD)
        try:
            if "@" in username:
                user = UserModel._default_manager.get(email=username)
            else:
                user = UserModel._default_manager.get_by_natural_key(username)
        except UserModel.DoesNotExist:
            UserModel().set_password(password)
        else:
            if user.check_password(password) and self.user_can_authenticate(user):
                return user
